import numpy
cimport numpy
import cython
cimport cython
from libc.stdlib cimport malloc, free, abort

cdef extern from 'c/kernel.c':
    ctypedef (void*) GSPHKernel
    GSPHKernel gaussian_kernel

cdef extern from 'c/sphrasterize.c':
    ctypedef struct GSPHImage "GSPHImage":
        int size[3]
        numpy.intp_t strides[3]
        void * data
        
    void gsph_image_init(GSPHImage * image, 
        char * dtype,
        int size[3],
        numpy.intp_t strides[3],
        void * data)
    void gsph_rasterize(GSPHImage * painter, GSPHKernel kernel,
        double pos[2], double sml, double * mvalue)

ctypedef fused floatingpos:
    cython.int [:, :]
    cython.long [:, :]
    cython.float [:, :]
    cython.double [:, :]

ctypedef fused floatingdata:
    cython.int [:, :]
    cython.float [:, :]
    cython.double [:, :]

ctypedef fused floatingsml:
    cython.float [:]
    cython.double [:]

ctypedef fused floatingimage:
    cython.float [:, :, :]
    cython.double [:, :, :]

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.overflowcheck(False)
@cython.nonecheck(False)
def paint(floatingpos pos, 
        floatingsml sml,
        floatingdata data,
        floatingimage image,
        object mask):
    """
        paint to image. Image must be of shape (nc, ny, nx)
        
    """
    cdef:
        GSPHImage img
        numpy.intp_t i
        int usemask
        numpy.ndarray[numpy.uint8_t, ndim=1, cast=True] maskbool
        int j
        int size[3]
        numpy.intp_t strides[3]
        GSPHKernel kernel


    assert data.shape[0] == image.shape[0]

    if mask is None:
        usemask = False
        maskbool = numpy.array([False])
    else:
        usemask = True
        maskbool = mask

        
    # translate the shape and strides 
    size[0] = image.shape[1]
    size[1] = image.shape[2]
    size[2] = image.shape[0] # number of channels
    strides[0] = image.strides[1]
    strides[1] = image.strides[2]
    strides[2] = image.strides[0]

    if image.itemsize == 8:
        gsph_image_init(&img, 'f8', size, strides, &image[0, 0, 0]);
    else:
        gsph_image_init(&img, 'f4', size, strides, &image[0, 0, 0]);

    kernel = gaussian_kernel

    cdef double * mvalue = <double*>malloc(size[2] * sizeof(double))
    cdef double dpos[2]
    cdef double dsml
    for i in range(0, pos.shape[0]):
        if usemask and not maskbool[i]: continue
        dpos[0] = pos[i, 0]
        dpos[1] = pos[i, 1]
        dsml = sml[i]
        for j in range(size[2]):
            mvalue[j] = data[j, i]
        gsph_rasterize(&img, kernel, dpos, dsml, mvalue)

    free(mvalue)



import numpy
from scipy.integrate import romberg
from scipy.interpolate import InterpolatedUnivariateSpline

# we keep cosmology.py relatively independent. Lazy is copied from
# lazy.py
class Lazy(object):
    def __init__(self, calculate_function):
        self._calculate = calculate_function

    def __get__(self, obj, _=None):
        if obj is None:
            return self
        value = self._calculate(obj)
        setattr(obj, self._calculate.func_name, value)
        return value

class interp1d(InterpolatedUnivariateSpline):
  """ this replaces the scipy interp1d which do not always
      pass through the points
      note that kind has to be an integer as it is actually
      a UnivariateSpline.
  """
  def __init__(self, x, y, kind, bounds_error=False, fill_value=numpy.nan, copy=True):
    if copy:
      self.x = x.copy()
      self.y = y.copy()
    else:
      self.x = x
      self.y = y
    InterpolatedUnivariateSpline.__init__(self, self.x, self.y, k=kind)
    self.xmin = self.x[0]
    self.xmax = self.x[-1]
    self.fill_value = fill_value
    self.bounds_error = bounds_error
  def __call__(self, x, nu=0):
    x = numpy.asarray(x)
    shape = x.shape
    x = x.ravel()
    bad = (x > self.xmax) | (x < self.xmin)
    if self.bounds_error and numpy.any(bad):
      raise ValueError("some values are out of bounds")
    y = InterpolatedUnivariateSpline.__call__(self, x.ravel(), nu=nu)
    y[bad] = self.fill_value
    return y.reshape(shape)

Np = 1000
logamin = -20
class Cosmology(object):
  """ every thing is dimensionless.
      Length is in DH = c / H0
  """

  def __init__(self, M, L, h, B, sigma8):
    self.M, self.L = M, L
    self.h = h
    self.sigma8 = sigma8
    self.B = B
  @Lazy
  def Dc(self):
    M, L = self.M, self.L
    logx = numpy.linspace(logamin, 0, Np)
    def kernel(loga):
      a = numpy.exp(loga)
      return 1 / self.Ea(a) * a ** -1 # dz = - 1 / a dloga
    y = numpy.array(
        [romberg(kernel, loga, 0, vec_func=True, divmax=10) for loga in logx])
    def func(xval,
            intp=interp1d(logx, y, kind=5)):
      return intp(numpy.log(xval))
    func.y = y
    func.x = numpy.exp(logx)
    func.__doc__ =  \
    """evaluates Dc(a) / DH
       (dimensionless, not multiplied by H0)
       for OmegaM = %g, OmegaL = %g
       .inv evaluates look-back time a from given Dc. 
    """ % (M, L)

    y = func.x[::-1]
    x = func.y[::-1]
    assert (x[1:] > x[:-1]).all()
    def inv(xval, nu=0,
            intp=interp1d(x, y, 
               bounds_error=True, 
               fill_value=numpy.nan, kind=5)):
      return intp(xval, nu=nu)
    inv.__doc__ = \
    """evaluates look-back time 
       (in terms of expansion factor a)
       at given dimensionless comoving distance
       (Dc/DH)
    """
    func.inv = inv
    return func

  @Lazy
  def aback(self):
    return self.Dc.inv

  @Lazy
  def Vrec(self):
    """ returns the recessing velocity at given time a,
        in unit of C"""
    x = self.Dc.x
    y = self.Dc.y

    v = y * self.Ea(x) * x

    def func(xval, nu=0,
            intp=interp1d(x, v, kind=5)):
      return intp(xval, nu)
    def inv(xval, nu=0,
            intp=interp1d(v[::-1], x[::-1], kind=5)):
      return intp(xval, nu)

    func.__doc__ = \
    """evaluates recessing velocity at time a
       (in terms of expansion factor a)
       in unit of C.
       .inv gives time a as function of recessing velocity
    """
    func.inv = inv
    return func

  def D(self, a):
    """ Dplus relative to z=0"""
    return self.Dplus(a) / self.Dplus(1.0)
  @Lazy
  def Dplus(self):
    M, L = self.M, self.L
    logx = numpy.linspace(logamin, 0, Np)
    x = numpy.exp(logx)
  
    def kernel(loga):
      a = numpy.exp(loga)
      return (a * self.Ea(a)) ** -3 * a # da = a * d loga
  
    y = self.Ea(x) * numpy.array(
                 [ romberg(kernel, logx.min(), loga, vec_func=True) 
                   for loga in logx])

    def func(x, nu=0, intp=interp1d(logx, y, 
            bounds_error=False, 
            fill_value=numpy.nan, kind=5)):
      return intp(numpy.log(x), nu=nu)
    func.__doc__ =  \
    """evaluates D+(a) There is no 2.5 factor neither.
       (dimensionless, not multiplied by H0)
       for OmegaM = %g, OmegaL = %g
    """ % (M, L)
    return func

  @Lazy
  def Ea(self):
    M, L = self.M, self.L
    func = lambda a: \
         a ** -1.5 * (M + (1 - M - L) * a + L * a ** 3) ** 0.5
    func.__doc__ = \
    """evaluates H(a)/H0 (dimensionless) 
       for OmegaM = %g, OmegaL = %g
    """ % (M, L)
    return func 

  @Lazy
  def FOmega(self):
    M, L = self.M, self.L
    def func(a):
      omega_a = M / (M + (1 - M - L) * a + L * a ** 3)
      return omega_a ** 0.6
      
    func.__doc__ = \
    """evaluates FOmega(a) 
       for OmegaM = %g, OmegaL = %g
    """ % (M, L)
    return func 

  @Lazy
  def Pk(self):
    import pycamb
    a = dict(H0=self.h * 100., 
          omegac=self.M- self.B, 
          omegab=self.B, 
          omegav=self.L, 
          omegak=1 - self.M - self.L, omegan=0)
    DH = 299792.468 / 100.
    fakesigma8 = pycamb.transfers(scalar_amp=1, **a)[2]
    scalar_amp = fakesigma8 ** -2 * self.sigma8 ** 2
    k, p = pycamb.matter_power(scalar_amp=scalar_amp, maxk=500, **a)
    k *= DH
    p /= DH ** 3
    p[numpy.isnan(p)] = 0
    func = interp1d(k, p, kind=5, bounds_error=False, fill_value=0)
    func.__doc__ = \
    """power spectrum P(k) normalized to sigma8.
       k is in 1/DH and p is in DH**3
    """
    return func

  def disp_to_vel(self, disp, a):
    """ convert init. displacement to init. velocity,
        time given by a
        losvel is physical / proper
        initial disp shall be in units of DH.
    """
    Ea = self.Ea
    Dplus = self.Dplus
    FOmega = self.FOmega
    D = Dplus(a) / Dplus(1.0)
    losvel = a * D * FOmega(a) * Ea(a) * disp
    return losvel

  def redshift_dist(self, losvel, a):
    """redshift distortion by los displacement
       for losvel, moving away is positive
       moving away is larger z, smaller a

       returns a, not z
    """
    # losvel is physical / proper
    # for losvel, moving away is posative
    # moving away is larger z
    #
    # from Martin White's 2012 Santa Fe notes
    Dc = self.Dc(a)
    Dcred = Dc + losvel / (a * self.Ea(a))
    ared = self.Dc.inv(Dcred)
    return ared

WMAP7 = Cosmology(M=0.272, L=0.728, h=0.702, B=0.046, sigma8=0.801)



"""
Survey Volume Remapping in Gaepsi

    3D transformation that remaps a cube to a sheet.

    need matrix M (hard to find, get the numbers from Carlson and White 2010

    remap expects input in [0, 1]

    wrap will convert BoxSize to [0, 1]
    remap_query_size returns the boxsize of the remapped volume 
            [1, 1, 1] -> size

"""

import numpy
cimport numpy

import cython
cimport cython

cdef extern from 'c/transform.c':
    pass
cdef extern from 'c/svremap.c':
    ctypedef struct SVRemap "SVRemap":
        double size[3]
        pass
    int svremap_init(SVRemap  * r, int * remap)
    double svremap_apply(SVRemap * r, double x[3], double y[3], int I[3])

def wrap(pos, BoxSize, offset=None, output=None):
    """ wrapping pos in a box of BoxSize into (0, 1)
        assuming the pos is of shape [..., Nd]
    """
    norm = numpy.empty(pos.shape[-1])
    norm[:] = BoxSize
    norm[:] = 1.0 / norm[:]
    if output is None:
        output = numpy.empty_like(pos)
    if offset and offset != 0.0 :
        off = numpy.empty(pos.shape[-1])
        off[:] = offset
        pos = numpy.substract(pos, off, output)
    numpy.multiply(pos, norm, output)
    return numpy.remainder(output, 1.0, output)

def remap_query_size(M):
    cdef SVRemap r 
    cdef int[:, ::1] matrix
    cdef double[::1] size
    matrix = numpy.array(M, dtype='int32')
    svremap_init(&r, &matrix[0, 0])
    return numpy.array(<double [:3]> r.size).copy()

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.overflowcheck(False)
@cython.nonecheck(False)
def remap(pos, M, output=None):
    """ the output if provided shall be of shape [, 3]
        pos shall be of shape [, 3], and between [0, 1] in all dimensions.
        M is a 3x3 integer matrix
        currently no sanity checks are performed.
    """
    if output is None: 
        output = numpy.empty_like(pos, dtype='f8')
    else:
        assert output.dtype == numpy.dtype('f8')

    assert output.shape[output.ndim - 1] == 3

    output[...] = pos

    cdef int[:, ::1] matrix
    cdef int I[3] 
    cdef double[:, ::1] x = output
    cdef double[:, ::1] y = output
    cdef double tmp[3]
    cdef SVRemap r 

    matrix = numpy.array(M, dtype='int32')
    I[0] = 0
    I[1] = 0
    I[2] = 0
    svremap_init(&r, &matrix[0, 0])
    cdef numpy.intp_t i
    for i in range(x.shape[0]):
        svremap_apply(&r, &x[i, 0], &tmp[0], I)
        y[i, 0] = tmp[0]
        y[i, 1] = tmp[1]
        y[i, 2] = tmp[2]
    return output




from . import _painter
import sharedmem
import numpy

def paint(pos, sml, data, shape, mask=None, np=0):
    """ Use SPH kernel to splat particles to an image.

        Parameters
        ----------

        pos : array_like
          (..., >2) position of particles. Only two first
          column is used. In device coordinate

        data : array_like
          (Nc, ...) or (...). Weight to use for painting.
          Nc channels will be produces on the device.
          if the array is 1d, Nc = 1
        
        sml : array_like
          smoothing length (half of effective size).
          In device coordinate; only correct in isotropic
          cameras

        shape : list, tuple
          (w[0], w[1]) the size of the device.
          shall enclose pos[..., 0] and pos[..., 1]

        mask : array_like, boolean
          If provided, elements with False will not be painted.

        np : int
          number of multiprocessing. 0 for single-processing.
          None for all available cores.

        Returns
        -------
        image: array_like
           (Nc, shape[0], shape[1])

        Notes
        -----
        Remember to transpose for imshow and pmesh to correct put x horizontaly.
    """

    if len(numpy.shape(data)) == 1:
        data = [data]

    with sharedmem.MapReduce(np=np) as pool:
        if pool.np > 0: nbuf = pool.np
        else: nbuf = 1
        buf = sharedmem.empty((nbuf, len(data)) + shape, dtype='f4')
        buf[:] = 0
        chunksize = 1024 * 8

        def work(i):
            sl = slice(i, i + chunksize)
            datas = [d[sl] for d in data]
            if mask is not None: masks = mask[sl]
            else: masks = None
            _painter.paint(pos[sl], sml[sl], numpy.array(datas),
                    buf[pool.local.rank], masks)

        pool.map(work, range(0, len(pos), chunksize))
    return numpy.sum(buf, axis=0)





"""
   Camera support in Gaepsi

   A camera is represented by a 4x4 transformation matrix like OpenGL.

   If it is done correctly, we use
   
   pos[:, :4] dot  matrix for the transformation.

   lookat: build the model/view matrix
   ortho:  the orthogonal perspective matrix

   the full transformation is ortho dot lookat

   apply: apply the transformation to a 3 d position vector,

   the devide coordinate is [-1, 1]. 

   data_to_device transforms pos and sml together.

"""

import numpy
import sharedmem

class projectionmatrix(numpy.ndarray):
    pass

class modelviewmatrix(numpy.ndarray):
    pass

class cameramatrix(numpy.ndarray):
    pass

def matrix(projection, modelview):
    """ Create a camera matrix.
    
        Parameters
        ----------
        projection: projectionmatrix
           from ortho or persp

        modelview : modelviewmatrix
           from lookat

        Returns
        -------
        cameramatrix
    """
    assert isinstance(projection, projectionmatrix)
    assert isinstance(modelview, modelviewmatrix)

    matrix = projection.dot(modelview).view(type=cameramatrix)
    matrix.near = projection.near
    matrix.far = projection.far
    matrix.up = modelview.up
    matrix.side = modelview.side

    return matrix

def ortho(near, far, extent):
    """ An orthogonal camera projection.
        
        Parameters
        ----------
        extent : tuple of 4
            left, right, top, bottom in data coordinate

        near : float
            location of the near field

        far : float
            location of the far field

        Returns
        -------
        A projectionmatrix.

    """
    l, r, b, t = extent
    ortho = numpy.zeros((4,4))
    ortho[0, 0] = 2.0 / (r - l)
    ortho[1, 1] = 2.0 / (t - b)
    ortho[2, 2] = -2.0 / (far - near)
    ortho[3, 3] = 1
    ortho[0, 3] = - (1. * r + l) / (r - l)
    ortho[1, 3] = - (1. * t + b) / (t - b)
    ortho[2, 3] = - (1. * far + near) / (far - near)
    ortho = ortho.view(type=projectionmatrix)
    ortho.extent = extent
    ortho.near = near
    ortho.far = far
    ortho.scale = numpy.array([ortho[0, 0], ortho[1, 1]])
    return ortho

def fov2extent(fov, aspect, D):
    """ Convert FOV parameters to extent """
    l = - numpy.tan(fov * 0.5) * aspect * D
    r = - l
    b = - numpy.tan(fov * 0.5) * D
    t = - b
    return (l, r, b, t)

def extent2fov(extent, D):
    """ Convert extent to FOV parameters """
    l, r, b, t = extent
    aspect = (l - r) /(b - t)
    fov = numpy.arctan2((t - b), D) * 2
    return fov, aspect

def persp(near, far, fov, aspect):
    """ An perspective camera projection.
        
        Parameters
        ----------
        near : float
            location of the near field

        far : float
            location of the far field

        fov : float
            Field of View angle in radians.

        aspect : float
            aspect ratio. It shall be the same as
            the device shape[0] / shape[1] for squared pixels.

        Returns
        -------
        A projectionmatrix.

        perspective dot modelview
    """
    
    # distance cancels out
    l, r, b, t = fov2extent(fov, aspect, 1)
    
    persp = numpy.zeros((4,4))
 
    persp[0, 0] = 2. / (r - l)
    persp[1, 1] = 2. / (t - b)
    persp[2, 2] = - (1. *(far + near)) / (far - near)
    persp[2, 3] = - (2. * far * near) / (far - near)
    persp[0, 2] = (r + l) / (r - l)
    persp[1, 2] = (t + b) / (t - b)
    persp[3, 2] = -1
    persp[3, 3] = 0

    persp = persp.view(type=projectionmatrix)
    persp.near = near
    persp.far = far
    persp.scale = numpy.array([persp[0, 0], persp[1, 1]])

    return persp

def lookat(pos, target, up):
    """ Set up a camera position matrix (modelview)

        Parameters
        ----------
        pos : 3-tuple
           Position of the camera

        target : 3-tuple
           Focal point of the camera

        up : 3-tuple
           Up direction of the camera

        Returns
        -------
        modelview matrix. The up and side vector are
        stored as attributes.

        Notes
        -----
        The full camera transformation is 

        m = dot(projection, modelview)

    """
    pos = numpy.asarray(pos)
    target = numpy.asarray(target)
    up = numpy.asarray(up)

    dir = target - pos
    dir = dir / (dir **2).sum() ** 0.5
    side = numpy.cross(dir, up)
    side[:] = side / (side **2).sum() ** 0.5
    up = numpy.cross(side, dir)
    up = up / (up **2).sum() ** 0.5

    m1 = numpy.zeros((4,4))
    m1[0, 0:3] = side
    m1[1, 0:3] = up
    m1[2, 0:3] = -dir
    m1[3, 3] = 1

    tran = numpy.eye(4)
    tran[0:3, 3] = -pos
    m2 = numpy.dot(m1, tran)
    m2 = m2.view(type=modelviewmatrix)
    m2.up = up
    m2.side = side
    return m2

def apply(matrix, pos, np=None):
    """ 
        apply a camera matrix to data coordinates

        Parameters
        ----------
        matrix : cameramatrix
           created by :func:`matrix`

        pos : array_like
           data cooridnates

        Returns
        -------
        clip_pos : array_like
           The position in the clip coordinate.
           The fustrum of the camera is in
           (-1, 1) x (-1, 1) x (-1, 1)

    """
    assert isinstance(matrix, cameramatrix)

    shmout = sharedmem.empty_like(pos)
    chunksize = 1024 * 32

    def work(i):
        tmppos = pos[i:chunksize+i]
        tmpout = shmout[i:chunksize+i]
        tmp = numpy.empty((len(tmppos), 4), dtype='f8')
        tmp[..., 3] = 1.0
        tmp[..., :3] = tmppos
        tmp = numpy.dot(tmp, matrix.T)

        tmpout[..., :] = tmp[..., :3] / tmp[..., 3][..., None]
        
    with sharedmem.MapReduce(np=np) as pool:
        pool.map(work, range(0, len(pos), chunksize))

    return shmout

def clip(xc):
    """ Compute the clipping mask from clipping coordinates.
    
        Parameters
        ----------
        xc : array_like
           position in clipping coordinate.
    
        Returns
        -------
        mask : array_like
           True for inside, False for outside.

    """

    return ((xc >= -1) & (xc <= 1)).all(axis=-1)

def todevice(xc, extent, np=None):
    """ Convert clipping coordinate to device coordinate.
    
        Parameters
        ----------
        extent : 2-tuple or 4-tuple
           The extent of the device cooridnate.
           (E[0], E[1]) or (S[0], E[0], S[1], E[1]).
           in first case S[0], and S[1] are 0.
           
        xc : array_like (... ,3)
            Clipping coordinates. Only the first two columns
            of xc are used.

        Returns
        -------
        xd : array_like (..., 2)
            data in device coordinate.
            For data in the fustrum, it shall be
            between S and E. Only the first two columns
            of xd are useful.

    """
    if len(extent) == 2:
        r, t = extent
        l, b = 0, 0
    else:
        l, r, b, t = extent

    chunksize = 1024 * 32
    out = sharedmem.empty((len(xc), 2))
    def work(i):
        tmp = (xc[i:i+chunksize] + 1.0)
        tmp *= 0.5
        tmp[..., 1] *= (t - b)
        tmp[..., 1] += b
        tmp[..., 0] *= (r - l)
        tmp[..., 0] += l
        out[i:i+chunksize] = tmp[:, :2]
    with sharedmem.MapReduce(np=np) as pool:
        pool.map(work, range(0, len(xc), chunksize))

    return out

def data_to_device(matrix, pos, sml, extent, apply_clip=False, np=None):
    """ Transform pos and smoothing length in device coordinate,
        to device with finite differentiation.
    
        Parameters
        ----------
        matrix: camera matrix

        pos : array_like
            position in data coordinate

        sml : array_like
            smoothing in data coordinate

        extent : 2-tuple or 4-tuple
            (see todevice)

        apply_clip : bool
            if True, the clipping will be applied
        Returns
        -------
        xd, smld, clip : array_like
            position and sml in device coordinate, and clipping flag.

            xd and smld are clipped by clip if apply_clip is True.

    """
    assert isinstance(matrix, cameramatrix)

    xc = apply(matrix, pos + sml[:, None] * matrix.side, np=np)

    m = clip(xc)
    if apply_clip:
        xc = xc[m]
        pos = pos[m]
        sml = sml[m]

    xd = todevice(xc, extent, np=np)

    xc1 = apply(matrix, pos + sml[:, None] * matrix.side, np=np)
    xd1 = todevice(xc1, extent, np=np)

    smld = ((xd1 - xd) ** 2).sum(axis=-1) ** 0.5
    return xd, smld, m

if __name__ == '__main__':
    test()



from setuptools import setup
from Cython.Build import cythonize
from distutils.extension import Extension
import numpy
def myext(*args):
    return Extension(*args, include_dirs=["./", numpy.get_include()],
            extra_compile_args=['-O3'],
            extra_link_args=['-O3']
            )

extensions = [
        myext("gaepsi2.svr", ["gaepsi2/svr.pyx"]),
        myext("gaepsi2._painter", ["gaepsi2/_painter.pyx"])
        ]


def find_version(path):
    import re
    # path shall be a plain ascii text file.
    s = open(path, 'rt').read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              s, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Version not found")

setup(
    name="gaepsi2", version=find_version('gaepsi2/version.py'),
    author="Yu Feng",
    author_email="rainwoodman@gmail.com",
    url="http://github.com/rainwoodman/gaepsi2",
    description="gaepsi2. SPH visualization.",
    install_requires=['cython', 'numpy', 'sharedmem'],
    zip_safe = False,
    packages= ['gaepsi2', 'gaepsi2.tests'],
    ext_modules = cythonize(extensions),
    license='BSD-2-Clause',
)




import numpy

def test_svr():
    from gaepsi2 import svr
    M=[[1, 1, 0], [0, 1, 0], [0, 0, 1]];
    print(svr.remap([[0, 0, 1.]], M))

def test_painter():
    from gaepsi2 import painter
    pos = numpy.array([
            numpy.arange(0, 1),
            numpy.arange(0, 1),
            ]).T
    print(pos)
    sml = numpy.ones(len(pos)) * 4
    data = numpy.ones(len(pos))
    print(painter.paint(pos, sml, [data], (5, 5)))
    print(painter.paint(pos + 0.5, sml, [data], (5, 5)))


def test_camera():
    from gaepsi2 import camera

    pos = numpy.random.uniform(size=(1000, 3)) * 20.
    pos -= 10.
    proj = camera.ortho(0, 20, (-10, 10, -10, 10))
    mv = camera.lookat((0, 0, -10), (0, 0, 0), (0, 1, 0))
    matrix = camera.matrix(proj, mv)
    p2d = camera.apply(matrix, pos)
